''' script for generating Luxrender lxs files for ToF Simulation'''

import re
import sys	
import math
import os
import subprocess
import h5py
import time
import uuid

#----------------------- Path to executables and workflow files ------------------------------------

luxconsole = "E:/Users/smeister/Luxrender/LuxSource_sppm/lux_build64/Release/luxconsole.exe"
tuchulcha_run = "E:/Users/smeister/Charon/installs/trunk/64/bin/tuchulcha-run.exe"
pack_workflow = "E:/Users/smeister/Scripts/genericWorkflows/TofSimulationPackage.wrp"
pack_ext_workflow = "E:/Users/smeister/Scripts/genericWorkflows/TofSimulationPackageExtended.wrp"
pack_gt_workflow = "E:/Users/smeister/Scripts/genericWorkflows/AddGt2H5.wrp"
results_outdir = "results"

#------------------------ Helper Functions ---------------------------------------------------------

def s2Int(s):
	if s[-1] in ["M","m"]:
		return int(s[:-1]) * 1000000
	if s[-1] in ["K","k"]:
		return int(s[:-1]) * 1000
	return int(s)

def int2S(i):
	if i // 1000000 > 0:
		return "%uM" % (i // 1000000)
	if i // 1000 > 0:
		return "%uk" % (i // 1000)
	return "%u" % i

def getVal(default):
	r = input().strip()
	if r == '':
		return default
	else:
		return r
	
def replaceOrInsert(value, section, target):
	if re.search(value,target):
		return target
	else:
		target = re.sub(section, '''%s\n\t%s''' % (section, value),target)
	return target

#---------------------------------------------------------------------------------------------------
help_txt = '''
This script takes multiple Luxrender lxs scene files as input and modifies them for ToF Simulation
You can either simulate 4 raw frames per scene file or only one.
In the later case the number of scene files must be divisible by 4. This is usefull e.g. if you want to render motion artifacts.

The following parameters can be adjusted:
eye depth: maximum length of eye path
light depth: maximum length of light path
	eye depth + light depth must be > 3 or the rendering will never converge
	to run a simulation without multipath set this to 2/1
	Otherwise for simple geometries values of 8/8 may be sufficient.
	If you have transparent materials, increase to 16 or higher
	The render time does only weakly depend on this value,
	it is advisable to choose a larger value if you are unsure
Samples: render samples per pixel
	You can supply values with suffixes (5k = 5000, 10M=10000000)
	The render time does scale linear with this value
	1k is usually more than enough for simple geometries without transparencies.
	More complex cases may need values of around 20k.
ToF Freq: ToF modulation frequency
	20Mhz is the default frequency for PMD CamCube cameras.
	This also controls the non-ambiguity range
Harmonics: Simulate wiggling error by adding higher harmonics to the light modulation
	leave at zero for no wiggling
Harmonics shift: phase shift between base frequency and higher harmonics
Harmonics Intensity: relative intensity between base frequency and higher harmonics
	settings this higher than 0.2 may produce strange results
Render output directory: directory where the rendered exr raw frames will be saved
lxs file prefix: string that will be put before each outputfile. Use this to organize multiple runs
cameramatrix file: txt file containing the ToF internal camera parameters
	File layout:
	fx, 0, cx
	0, fy, cy
	0, 0, 1
	
	(note the extra blank line at the end!)
	fx,fy = focal length in pixel
	cx,cy = principal point in pixel
	
	The values can be computed from the settings inside blender this way
	fx,fy = (resolution [px] / 2.0) / tan(Field-of-View / 2.0)
	cx, cy = resultion [px] / 2.0 + 0.5
	
comments: comment that will be added to the output hdf5 as attribute

Send questions to Stephan Meister, HCI,IWR, University of Heidelberg
<stephan.meister@iwr.uni-heidelberg.de>
'''
	
#------------------------------------------- Main Script -------------------------------------------
	
if __name__ == "__main__":
	if len(sys.argv) == 2 and sys.argv[1] == "--help" :
		print(help_txt)
		exit()
	if len(sys.argv) < 5:
		print("Call with <lxs file format string> <ground truth format string> firstIndex endIndex [increment]!")
		print("e.g.: python3 simulateTofSequence.py myfile%04u.lxs 0 3 1")
		print("This will use the files myfile0000.lxs myfile0001.lxs and myfile0002.lxs")
		print("Or Call with --help for more information")
		exit()
	increment = 1
	if len(sys.argv) >= 6:
		increment = int(sys.argv[5])
	t1 = time.clock()
	sceneFiles = []
	gtFiles = []
	print("Using scene files:")
	for i in range(	int(sys.argv[3]),int(sys.argv[4]),increment):
		sceneFiles.append((sys.argv[1] % i).replace("\\","/"))
		gtFiles.append((sys.argv[2] %i).replace("\\","/"))
		print(sceneFiles[-1])


	# --------------------------------- lxs file template ------------------------
	print("Generate four phase images per lxs_base file? (No motion artifacts) (Y/N) [Y]")
	tmp = getVal("Y")
	fourRendersPerLxsBase = True
	if tmp in "YyJj1":
		fourRendersPerLxsBase = True
	elif tmp in "nN0":
		if len(sceneFiles) % 4 != 0:
			print("In this case the number of scene files must be divisible by 4!")
			exit()
		fourRendersPerLxsBase = False
	else:
		print("Use n,N,0 or y,Y,j,J,1 as input")
		exit()
	
	print("Eye depth [6]: ", end="")
	eyeDepth = s2Int(getVal("6"))
	print("Light depth [6]: ", end="")
	lightDepth = s2Int(getVal("6"))
	print("Samples [1k]: ",end="")
	samples = s2Int(getVal("1k"))
	print("ToF Freq [20M]: ",end="")
	toffreq = s2Int(getVal("20M"))
	print("Harmonics [0]: ", end="")
	harmonics = s2Int(getVal("0"))
	print("Harmonics shift[0.0]: ", end="")
	harmonicsShift = (getVal("0.0"))
	print("Harmonics Intensity[0.0]: ", end="")
	harmonicsInt = (getVal("0.0"))

	print("Render output directory [render]: ",end="")
	render_outdir = getVal("render").replace('''\\''','''/''')
	print("lxs file prefix [render]: ",end="")
	file_prefix = getVal("render")
	print("cameramatrix file [../../tof_internal.txt]:",end="")
	cam_matrix = getVal("../../tof_internal.txt")
	print("comments []",end="")
	comments = getVal("")
	
	if not render_outdir == "" and not os.path.exists(render_outdir):
		os.makedirs(render_outdir)

	if not os.path.exists("tmp"):
		os.makedirs("tmp")
		
	#-------------------------------- create hdf5 basefile-----------------------------------------------
	
	if not os.path.exists(results_outdir):
		os.makedirs(results_outdir)
	h5_filename = "%s/%s_%seyeDepth_%slightDepth_%sSamples.h5" % (results_outdir, file_prefix, int2S(eyeDepth), int2S(lightDepth), int2S(samples))
	
	if os.path.exists(h5_filename):
		print("experiment result file already exists! Overwrite? (y/n) [n]",end="")
		leave = getVal("n")
		if leave != "y":
			exit()
	
	h5f = h5py.File(h5_filename)
	
	root = h5f["/"]
	root.attrs["comments"] = comments
	root.attrs["scene files"] = "\n".join(sceneFiles)
	h5f.close()
	#---------------------------------------------------------------
	index = 0
	phaseshifts = [0, 0.5 * math.pi, 1.0 * math.pi, 1.5 * math.pi]


	t1 = time.clock()
	lxsFiles = []
	exr_files = []

	for lxs_base in sceneFiles:
		f = open(lxs_base)
		template = f.read()
		f.close()
	
	
		p = re.compile(r'''eyedepth" \[[0-9]+\]''')
		template = re.sub(p,'''eyedepth" [%u]''' % eyeDepth,template,count = 1)
		p = re.compile(r'''lightdepth" \[[0-9]+\]''')
		template = re.sub(p,'''lightdepth" [%u]''' % lightDepth,template,count = 1)
		p = re.compile(r'''haltspp" \[[0-9]+\]''')
		template = re.sub(p,'''haltspp" [%u]''' % samples,template,count = 1)

		p = re.compile('''"integer displayinterval" \[[0-9]+\]''')
		template = re.sub(p,'''"integer displayinterval" [60]''',template,count = 1)
		p = re.compile('''"integer writeinterval" \[[0-9]+\]''')
		template = re.sub(p,'''"integer writeinterval" [60]''',template,count = 1)
		p = re.compile('''"integer flmwriteinterval" \[[0-9]+\]''')
		template = re.sub(p,'''"integer flmwriteinterval" [120]''',template,count = 1)

		template = replaceOrInsert('''"float toffreq" [%s]''' % toffreq,'''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"float modint" [0.75]''','''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"float modoffset" [0.25]''','''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"float phaseshift" [0]''','''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"integer harmonics" [%s]'''% harmonics,'''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"float harmonicsint" [%s]'''% harmonicsInt,'''SurfaceIntegrator "bidirectional"''',template)
		template = replaceOrInsert('''"float harmonicsshift" [%s]'''% harmonicsShift,'''SurfaceIntegrator "bidirectional"''',template)
	
		#-----------------------------
		
		if fourRendersPerLxsBase:
			
			for i in [0,1,2,3]:
				p = re.compile(r'''filename" \[".*?"\]''')
				exr_filepath = "%s/%s_%s_%seyeDepth_%slightDepth_%sSamples_%u" % (render_outdir, file_prefix,index,int2S(eyeDepth),int2S(lightDepth),int2S(samples),i)
				exr_files.append(exr_filepath+".exr")
				template = re.sub(p,'''filename" ["%s"]''' %(exr_filepath), template,count = 1)
				
				p = re.compile(r'''phaseshift" \[.*?]''')
				template = re.sub(p,'''phaseshift" [%f]''' % phaseshifts[i],template,count = 1)
				
				filename = "%s_%s_%seyeDepth_%slightDepth_%sSamples_%u.lxs" % (file_prefix, index,int2S(eyeDepth), int2S(lightDepth), int2S(samples),i)
				out = open(filename,"w")
				out.write(template)
				out.close()
				lxsFiles.append(filename)
				
		else:
				i = index % 4
				p = re.compile(r'''filename" \[".*?"\]''')
				exr_filepath = "%s/%s_%s_%seyeDepth_%slightDepth_%sSamples_%u" % (render_outdir, file_prefix,index,int2S(eyeDepth),int2S(lightDepth),int2S(samples),i)
				exr_files.append(exr_filepath+".exr")
				template = re.sub(p,'''filename" ["%s"]''' %(exr_filepath), template,count = 1)
				
				p = re.compile(r'''phaseshift" \[.*?]''')
				template = re.sub(p,'''phaseshift" [%f]''' % phaseshifts[i],template,count = 1)
				
				filename = "%s_%s_%seyeDepth_%slightDepth_%sSamples_%u.lxs" % (file_prefix, index,int2S(eyeDepth), int2S(lightDepth), int2S(samples),i)
				out = open(filename,"w")
				out.write(template)
				out.close()
				lxsFiles.append(filename)
		index += 1	
	
	index = 0
	for i in range(0,len(lxsFiles) - 3,4):
	
		processes = []
		logfiles =[]
		t2 = time.clock()
		print("Starting renderers")
		for j  in [0,1,2,3]:
			logfiles.append(open("tmp/%s" % (lxsFiles[i+j].replace(".lxs",".log")),"w"))
			p = subprocess.Popen([luxconsole, lxsFiles[i+j]],bufsize=-1, stdout=logfiles[-1],stderr=subprocess.STDOUT,stdin=subprocess.PIPE)
			processes.append(p)
			processes[-1].stdin.close()
		for p in processes:
			p.wait()
		for f in logfiles:
			f.close()
			
		print("Luxrender runs finished")
		print("Render Time: %.1f sec" % (time.clock() -t2))
			
		f = None
		if i == 0:
			f = open(pack_workflow)
		else:
			f = open(pack_ext_workflow)
		template = f.read()
		f.close()

		template = template.replace("$input$",";".join(exr_files[i:i+4]))
		template = template.replace("$output$",h5_filename)
		template = template.replace("$cmat$",cam_matrix)
		template = template.replace("$toffreq$",str(toffreq))
		
		pack_out_filename = "tmp_workflow_" + str(uuid.uuid4())[0:7] + ".wrp"
		out = open(pack_out_filename,"w")
		out.write(template)
		out.close()
	
		print("Running charon workflow")
		tlog = open("tmp/%s_tuchulcha.log" % lxsFiles[i][:-4],"w")
		p = subprocess.Popen([tuchulcha_run, "--non-interactive","--quiet","run",pack_out_filename],bufsize=-1, stdout=tlog,stderr=subprocess.STDOUT,stdin=subprocess.PIPE)
		p.stdin.close()
		p.wait()
		os.remove(pack_out_filename)
		index += 1
		print("\n")
	
	f = open(pack_gt_workflow)
	template = f.read()
	f.close()
	
	template = template.replace("$input$",";".join(gtFiles))
	template = template.replace("$output$",h5_filename)
	template = template.replace("$cmat$",cam_matrix)
	
	pack_out_filename = "tmp_workflow_" + str(uuid.uuid4())[0:7] + ".wrp"
	out = open(pack_out_filename,"w")
	out.write(template)
	out.close()
	
	tlog = open("tmp/gt_tuchulcha.log" ,"w")
	p = subprocess.Popen([tuchulcha_run, "--non-interactive","--quiet","run",pack_out_filename],bufsize=-1, stdout=tlog,stderr=subprocess.STDOUT,stdin=subprocess.PIPE)
	p.stdin.close()
	p.wait()
	os.remove(pack_out_filename)
	tlog.close()
	#--------------------------- convert data ------------------------
		
	print("\nfinished")
	print("Total Time: %.1f sec" % (time.clock() -t1))
	