/*  Copyright (C) 2010 Stephan Meister

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FileReaderExr.hxx
 *  Implementation of the parameter class FileReaderExr.
 *  \author Stephan Meister
 *  \date 23.03.2012
 */

#ifndef _FILEREADEREXR_HXX_
#define _FILEREADEREXR_HXX_

#include "FileReaderExr.h"
#include <fstream>
#include <sstream>
#include <boost/smart_ptr.hpp>
#include <boost/algorithm/string.hpp>

#include <ImfRgbaFile.h>
#include <ImfOutputFile.h>
#include <ImfInputFile.h>
#include <ImfChannelList.h>
#include <ImfArray.h>

using namespace Imf;
using namespace Imath;
using namespace cimg_library;

using namespace cimg_library ;

void readRGBA1 (const char fileName[],
				Array2D<float> &rPixels,
				Array2D<float> &gPixels,
				Array2D<float> &bPixels,
				Array2D<float> &aPixels,
				int &width, int &height)
{
	InputFile file (fileName);
	Box2i dw = file.header().dataWindow();
	width  = dw.max.x - dw.min.x + 1;
	height = dw.max.y - dw.min.y + 1;
	rPixels.resizeErase (height, width);
	gPixels.resizeErase (height, width);
	bPixels.resizeErase (height, width);
	aPixels.resizeErase (height, width);

	FrameBuffer frameBuffer;
	frameBuffer.insert ("R",                                  // name
		Slice (Imf::FLOAT,                          // type 
				(char *) (&rPixels[0][0] -     // base
							dw.min.x -
							dw.min.y * width),
				sizeof (rPixels[0][0]) * 1,    // xStride
				sizeof (rPixels[0][0]) * width,// yStride
				1, 1,                          // x/y sampling
				0.0));                         // fillValue

	frameBuffer.insert ("G",                                  // name
		Slice (Imf::FLOAT,                          // type
				(char *) (&gPixels[0][0] -     // base
							dw.min.x -
							dw.min.y * width),
				sizeof (gPixels[0][0]) * 1,    // xStride
				sizeof (gPixels[0][0]) * width,// yStride
				1, 1,                          // x/y sampling
				0.0));                         // fillValue

	frameBuffer.insert ("B",                                  // name
		Slice (Imf::FLOAT,                         // type
				(char *) (&bPixels[0][0] -     // base
							dw.min.x -
							dw.min.y * width),
				sizeof (bPixels[0][0]) * 1,    // xStride
				sizeof (bPixels[0][0]) * width,// yStride
				1, 1,                          // x/y sampling
				0.0));                     // fillValue
		
	frameBuffer.insert ("A",                                  // name
		Slice (Imf::FLOAT,                          // type
				(char *) (&aPixels[0][0] -     // base
							dw.min.x -
							dw.min.y * width),
				sizeof (aPixels[0][0]) * 1,    // xStride
				sizeof (aPixels[0][0]) * width,// yStride
				1, 1,                          // x/y sampling
				0.0));                         // fillValue
		
	file.setFrameBuffer (frameBuffer);
	file.readPixels (dw.min.y, dw.max.y);
}

template <typename T>
FileReaderExr<T>::FileReaderExr(const std::string& name) :
		TemplatedParameteredObject<T>("FileReaderExr", name, 
			"Read OpenEXR image files<br>"
			"multiple files will be concatenated in a CImgList<br>"
			"Color channels will be saved in the t(spectrum) dimension")
{
	ParameteredObject::_setTags("CoudCollection;DiskIO") ;

	ParameteredObject::_addParameter( _filenames, "filenames", "input filenames, separated by \";\"") ;

	ParameteredObject::_addParameter( _channels, "channels", "channels to read from exr", "{R;G;B;A;RGB;RGBA}") ;

	ParameteredObject::_addOutputSlot(_out, "in", "output image", "CImgList<T>"); 
}

template <typename T>
void FileReaderExr<T>::execute() {
	using namespace std ;
	
	std::vector<std::string> strs;
	boost::split(strs, _filenames(), boost::is_any_of(";"));

	if(strs.size() < 1)
	{
		ParameteredObject::raise("No filenames defined!") ;
	}
	
	Array2D<float> r, g, b,a ;
	int width = 0 ;
	int height = 0 ;
	try
	{	readRGBA1(strs[0].c_str(), r, g, b, a,width, height) ;
	}
	catch(std::exception& err)
	{	ParameteredObject::raise(err.what()) ;	}
	
	int channels = (_channels() == "RGBA" ? 4 : (_channels() == "RGB" ? 3 : 1)) ;
	sout << "Generating " << width << " x " << height << " x 1 x " << channels << " x " << strs.size() << " image" << std::endl ;
	_out().assign(strs.size(),width, height, 1, channels) ;
	for(size_t ii = 0 ; ii < strs.size() ; ii++)
	{
		sout << "Reading image # " << ii << std::endl ;
		int width_ = 0 ;
		int height_ = 0 ;

		try
		{	readRGBA1(strs[ii].c_str(), r, g, b, a,width_, height_) ;
		}
		catch(std::exception& err)
		{	ParameteredObject::raise(err.what()) ;	}
		if(width != width_ || height != height_)
		{	ParameteredObject::raise(std::string("Dimensions of img \"") + strs[ii] + "\" do not match previous images!") ;	}

		CImg<T>& img = _out()[ii] ;
		if(_channels() == "RGBA")
		{
			std::copy(r[0],r[0]+width*height,&img(0,0,0,0)) ;
			std::copy(g[0],g[0]+width*height,&img(0,0,0,1)) ;
			std::copy(b[0],b[0]+width*height,&img(0,0,0,2)) ;
			std::copy(a[0],a[0]+width*height,&img(0,0,0,3)) ;
		}
		else if(_channels() == "RGB")
		{
			std::copy(r[0],r[0]+width*height,&img(0,0,0,0)) ;
			std::copy(g[0],g[0]+width*height,&img(0,0,0,1)) ;
			std::copy(b[0],b[0]+width*height,&img(0,0,0,2)) ;
		}
		else if(_channels() == "R")
		{
			std::copy(r[0],r[0]+width*height,&img(0,0,0,0)) ;
		}
		else if(_channels() == "G")
		{
			std::copy(g[0],r[0]+width*height,&img(0,0,0,0)) ;
		}
		else if(_channels() == "B")
		{
			std::copy(b[0],r[0]+width*height,&img(0,0,0,0)) ;
		}

	}
}

#endif /* _DIST2DEPTH_HXX_ */



