/*  Copyright (C) 2012 Stephan Meister

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file RadialZConversion.hxx
 *  Implementation of the parameter class RadialZConversion.
 *  \author <a href="mailto:stephan.meister@iwr.uni-heidelberg.de">
 *      Stephan Meister</a>
 *  \date 30.05.2012
 */

#ifndef _RadialZConversion_HXX_
#define _RadialZConversion_HXX_

#define _USE_MATH_DEFINES
#include <cmath>
#include <charon-core/ParameteredObject.hxx>
#include "RadialZConversion.h"
#include <gsl/gsl_multifit.h>


template <typename T>
RadialZConversion<T>::RadialZConversion(const std::string& name) :
		TemplatedParameteredObject<T>(
			"RadialZConversion", name,
			"Convert z depth to radial depth and vice versa"
			
		),
		zDepthIn(true, false),
		radialDepthIn(true, false)
{
	ParameteredObject::_setTags("CoudCollection/Camera") ;
	ParameteredObject::_addInputSlot(
		zDepthIn, "zDepthIn",
		"input z depth",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		radialDepthIn, "radialDepthIn",
		"input radial depth",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		camMatrix, "camMatrix",
		"cameraMatrix (when not supported, view angles must be provided)",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		zDepthOut, "zDepthOut",
		"output z depth",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		radialDepthOut, "radialDepthOut",
		"output radial depth",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		xOut, "xOut",
		"output X coodinates",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		yOut, "yOut",
		"output Y coordinates",
		"CImgList<T>");

}

template <typename T>
void RadialZConversion<T>::execute() {

	using namespace cimg_library;
	
	if(zDepthIn.connected())
	{
		double fx = 0.0 ;
		double fy = 0.0 ;
		double cx = 0.0 ;
		double cy = 0.0 ;
		radialDepthOut().assign(zDepthIn()) ;
		xOut().assign(zDepthIn());
		yOut().assign(zDepthIn());
		{
			const CImg<T>& cm = camMatrix()[0] ;
			if(cm.width() < 3 || cm.height() < 3)
				ParameteredObject::raise("camera matrix has wrong dimensions") ;
			fx = cm(0,0) ; fy = cm(1,1) ;
			cx = cm(2,0) ; cy = cm(2,1) ;
			if(fx <= 0.0 || fy <= 0.0)
				ParameteredObject::raise("illegal focal lengths in cam matrix!") ;

			cimglist_for(zDepthIn(),l)
			{
				const CImg<T>& z = zDepthIn()[l] ;
				CImg<T>& rad = radialDepthOut()[l] ;
				CImg<T>& xout = xOut()[l];
				CImg<T>& yout = yOut()[l];
				//calculate radial distance from z-depth and lens view angle
				cimg_forXYZC(rad,xx,yy,zz,cc)
				{
					double Z = z(xx,yy,zz,cc) ;
					double X = (xx - cx) / fx * Z ;
					double Y = (yy - cy) / fy * Z ;
					rad(xx,yy,zz,cc) = T(sqrt(X*X + Y*Y + Z*Z)) ;
					xout(xx,yy,zz,cc) = T(X) ;
					yout(xx,yy,zz,cc) = T(Y) ;
				}
			}
		}
	}

	if(radialDepthIn.connected())
	{
		double fx = 0.0 ;
		double fy = 0.0 ;
		double cx = 0.0 ;
		double cy = 0.0 ;

		zDepthOut().assign(radialDepthIn()) ;
		xOut().assign(radialDepthIn());
		yOut().assign(radialDepthIn());
		{
			const CImg<T>& cm = camMatrix()[0] ;
			if(cm.width() < 3 || cm.height() < 3)
				ParameteredObject::raise("camera matrix has wrong dimensions") ;
			fx = cm(0,0) ; fy = cm(1,1) ;
			cx = cm(2,0) ; cy = cm(2,1) ;
			if(fx <= 0.0 || fy <= 0.0)
				ParameteredObject::raise("illegal focal lengths in cam matrix!") ;

			cimglist_for(radialDepthIn(),l)
			{
				const CImg<T>& R = radialDepthIn()[l] ;
				CImg<T>& Z = zDepthOut()[l] ;
				CImg<T>& xout = xOut()[l];
				CImg<T>& yout = yOut()[l];

				//calculate radial distance from z-depth and lens view angle
				cimg_forXYZC(R,xx,yy,zz,cc)
				{
					double r = R(xx,yy,zz,cc) ;
					double x = (xx - cx) / fx ;
					double y = (yy - cy) / fy ;
					//double X = x / (sqrt(x*x + fx * fx)) * r ;
					//double Y = y / (sqrt(y*y + fy * fy)) * r ;
					double z = r / sqrt (x*x + y*y + 1) ;
					x *= z ;
					y *= z ;
					Z(xx,yy,zz,cc) = T(z) ;
					xout(xx,yy,zz,cc) = x ;
					yout(xx,yy,zz,cc) = y ;
				}
			}
		}
	}
}

#endif /* _RadialZConversion_HXX_ */
