/*  Copyright (C) 2012 Stephan Meister

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file TofSim.hxx
 *  Implementation of the parameter class TofSim.
 *  \author <a href="mailto:stephan.meister@iwr.uni-heidelberg.de">
 *      Stephan Meister</a>
 *  \date 30.05.2012
 */

#ifndef _TOFSIM_HXX_
#define _TOFSIM_HXX_

#define _USE_MATH_DEFINES
#include <cmath>
#include <charon-core/ParameteredObject.hxx>
#include "TofSim.h"


template <typename T>
TofSim<T>::TofSim(const std::string& name) :
		TemplatedParameteredObject<T>(
			"TofSim", name,
			"<h2>Calculated depth, Intensity and Amplitude from raw tof "
			"data</h2><br>"
			"Calculated depth, Intensity and Amplitude from raw tof data"
		)
{
	ParameteredObject::_setTags("CoudCollection") ;
	
	ParameteredObject::_addInputSlot(
		phase0, "phase0",
		"Input for all phase shift images ordered in dimension 4",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		intensity, "int",
		"Intensity",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		amp, "amp",
		"Amplitude",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		depth, "depth",
		"Depth",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		phase, "phase",
		"Phase",
		"CImgList<T>");

	ParameteredObject::_addParameter< double >(
		frequency, "frequency",
		"tof modulation frequency",
		20e6, "double");
	ParameteredObject::_addParameter< double >(
		ambOffset, "ambiguityOffset",
		"add (2 * c / frequency) times this value to the depth",
		0.0, "double") ;
	ParameteredObject::_addParameter< double >(
		cubicCoef, "cubicCoefficient",
		"coefficient for cubic non-linearity on raw channels",
		0.0, "double") ;


	ParameteredObject::_addParameter(
		noiseDistribution, "noiseDistribution",
		"Distribution to sample from",
		"{none;gaussian;uniform;poisson}") ;
	ParameteredObject::_addParameter<float>(
		relNoiseAmplitude, "noiseAmplitude",
		"Noise Amplitude",
		1.0) ;
}

template <typename T>
void TofSim<T>::execute() {

	using namespace cimg_library;

	if(phase0().width() < 4)
	{
		ParameteredObject::raise("phase0 must be of size 4!") ;
	}
	
	CImg<T> I0 = phase0()[0] ;
	CImg<T> I1 = phase0()[1] ;
	CImg<T> I2 = phase0()[2] ;
	CImg<T> I3 = phase0()[3] ;

	if(cubicCoef != 0.0)
	{
		I0 += cubicCoef() * I0.get_pow(2.0) ;
		I1 += cubicCoef() * I1.get_pow(2.0) ;
		I2 += cubicCoef() * I2.get_pow(2.0) ;
		I3 += cubicCoef() * I3.get_pow(2.0) ;
	}

	double noise = relNoiseAmplitude() ;
	int noiseType = -1 ;
	if(noiseDistribution() == "gaussian")
	{
		noiseType = 0 ;
	}
	else if(noiseDistribution() == "uniform")
	{
		noiseType = 1 ;
	}
	else if(noiseDistribution() == "poisson")
	{
		noiseType = 3 ;
	}
	if(noiseType >= 0)
	{
		cimg_forXY(I0,xx,yy)
		{
			I0(xx,yy) += fabs(cimg::grand() * sqrt(double(I0(xx,yy))) * noise) ;
			I1(xx,yy) += fabs(cimg::grand() * sqrt(double(I1(xx,yy))) * noise) ;
			I2(xx,yy) += fabs(cimg::grand() * sqrt(double(I2(xx,yy))) * noise) ;
			I3(xx,yy) += fabs(cimg::grand() * sqrt(double(I3(xx,yy))) * noise) ;
		}
	}

	
	intensity().assign(I0) ;
	intensity()[0] = (I0 + I1 + I2 + I3) / 4.0 ;
	amp().assign(I0) ;
	CImg<T>& a = amp()[0] ;
	a = (I0 - I2).get_pow(2.0) + (I3 - I1).get_pow(2.0) ;
	a = a.get_sqrt() * M_PI / 2.0 ;

	phase().assign(I0) ;
	CImg<T>& p = phase()[0] ;
	CImg<T> p2 = phase()[0] ;
	
	p = (I3 - I1) ;
	p2 = (I0-I2) ;
	p.atan2(p2) ;
	
	/*
	p = (I3 -I1) ;
	cimg_forXY(p,x,y) 
	{	double v = I2(x,y)-I0(x,y) ;
		p(x,y) = (fabs(v) > 1e-6 ? p(x,y) / v : 0.0) ;
	}
	p.atan() ;
	*/

	//resulting phase is in range [-pi:pi]
	//the depths are mapped from this phase range [0:pi][-pi:0]
	
	depth().assign(I0) ;
	//depth()[0] = ((p / M_PI) + ambOffset()*2.0)  * (3e8 / frequency()) ;
	
	depth()[0] = p ;
	CImg<T>& d = depth()[0] ;
	cimg_forXYZC(p,x,y,z,c)
	{
		T v = p(x,y,z,c) ;
		d(x,y,z,c) = (v < 0 ? v + (2.0* M_PI) : v) ;
	}
	d = ((d / (2.0 * M_PI)) + ambOffset() ) * 0.5 * (3e8 / frequency()) ;
	//depth()[0] = ((p / (2.0 * M_PI)) + 0.5 + ambOffset())  * (6e8 / frequency()) ;
}

#endif /* _TOFSIM_HXX_ */
